#Region "FileHeader"
#If False Then
Delay-Mechanismus, mit dem eine Action auf den Zeitpunkt Application.EnterIdle verschoben werden kann. Der Vorgang kann vielfach angestoen werden - aufgerufen wird die Action nur einmal.
Ausgefhrt in 2 Varianten:
1) als WithEvents-Objekt. .SetRunning(true/false) lst das BeforeBusyEnds-Event - verzgert - aus, oder bricht auch ab. Bischen umstndlich zu coden, aber optimale Performance auch bei massenhaften  .SetRunning() - Mehrfach-Calls

2) als Shared Methode: Minimaler Programmier-Aufwand - man bergibt einfach einen AddressOf-Delegaten. 
Vorsicht bei anonymen Delegaten - sind das verschiedene Instanzen, werden sie alle dem _MultiCallback zugefgt, und die unerwnschte Massen-Ausfhrung nur "vertagt".

Benchmarks (Massenaufrufe pro Sekunde):
Leerlauf:            70175089
BusyDelay-Instanz:   33697961
Shared SetDelay:      6223927
Das Withevents-Verfahren liegt also an der Grenze der Messbarkeit, die Test-Schleife frisst ungefhr soviel Performance wie das getestete selbst. Daher tte ich als bereinigten Wert mindestens das doppelte annehmen.
Somit ist die Shared-Version 10-fach langsamer - aber hey! - 6 Mio Aufrufe die Sekunde - das wird nie der Flaschenhals sein.
#End If '-- Options, Imports
#End Region 'FileHeader


Imports System.ComponentModel
Imports System.Diagnostics

Namespace System.Windows.Forms

	Public Class BusyDelay

		Private _IsActive As Boolean

		Public Event BeforeBusyEnds As EventHandler

		''' <summary> (de-)aktiviert das Warten auf den Eintritt in die Leerlauf-Warteschleife. </summary>
		''' <remarks>Mehrfachaufrufe werden ignoriert</remarks>
		Public Sub SetRunning(Value As Boolean)
			If _IsActive = Value Then Return
			_IsActive = Value
			If _IsActive Then
				AddHandler Application.Idle, AddressOf Application_Idle
			Else
				RemoveHandler Application.Idle, AddressOf Application_Idle
			End If
		End Sub

		Private Sub Application_Idle(sender As Object, e As EventArgs)
			RemoveHandler Application.Idle, AddressOf Application_Idle
			_IsActive = False
			RaiseEvent BeforeBusyEnds(Me, e)
		End Sub

		<DebuggerStepThrough>
		Private Shared Sub BusyDelayCore_BeforeBusyEnds(sender As Object, e As EventArgs)
			If _MultiCallback.Null Then Return
			Dim act = DirectCast(_MultiCallback, Action)
			_MultiCallback = Nothing
			act()
		End Sub

		Private Shared _MultiCallback As [Delegate]

		Public Shared Sub SetCallback(cb As Action)
			' erst removen, dann wieder setzen
			_MultiCallback = [Delegate].Combine([Delegate].Remove(_MultiCallback, cb), cb)
		End Sub
		Public Shared Sub RemoveCallback(cb As Action)
			_MultiCallback = [Delegate].Remove(_MultiCallback, cb)
		End Sub
		Shared Sub New()
			AddHandler Application.Idle, AddressOf BusyDelayCore_BeforeBusyEnds
		End Sub

	End Class 'BusyDelay

End Namespace